import unittest
import os
from tempfile import mkdtemp
import subprocess
import pylha
from shutil import rmtree


my_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(my_path, '..', 'data')


class TestSMEFTsim(unittest.TestCase):
    def test_wcxf2smeftsim(self):
        tmpdir = mkdtemp()
        # use default settings
        res = subprocess.run(['wcxf2smeftsim',
                              os.path.join(data_path, 'test.Warsaw_mass.yml')],
                             cwd=tmpdir)
        # check return code
        self.assertEqual(res.returncode, 0, msg="Command failed")
        # check if file is present
        outf = os.path.join(tmpdir, 'wcxf2smeftsim_param_card.dat')
        self.assertTrue(os.path.isfile, outf)
        # check if can be imported as LHA
        with open(outf, 'r') as f:
            out = pylha.load(f)
        # check string is not empty
        self.assertTrue(out)
        # remove tmpdir
        rmtree(tmpdir)



import unittest
import numpy as np
import numpy.testing as npt
import yaml
import json
import pkgutil
import wcxf
import subprocess
import tempfile
import os

def _del_files(fs):
    """Delete files if they exist"""
    for f in fs:
        try:
            os.remove(f)
        except OSError:
            pass

class TestCLI(unittest.TestCase):
    def test_convert(self):
        yml1 = pkgutil.get_data('wcxf', 'data/test.basis1.yml').decode('utf-8')
        d_yml1 = yaml.load(yml1)
        # YAML stdin -> JSON stdout
        res = subprocess.run(['wcxf', 'convert', 'json', '-'],
                             input=yml1.encode(),
                             stdout=subprocess.PIPE)
        json1 = res.stdout.decode('utf-8')
        d_json1 = json.loads(json1)
        self.assertDictEqual(d_yml1, d_json1)
        # JSON stdin -> YAML stdout
        res = subprocess.run(['wcxf', 'convert', 'yaml', '-'],
                             input=json1.encode(),
                             stdout=subprocess.PIPE)
        yml2 = res.stdout.decode('utf-8')
        d_yml2 = yaml.load(yml2)
        self.assertDictEqual(d_json1, d_yml2)
        # YAML file -> JSON file
        _, fin = tempfile.mkstemp()
        with open(fin, 'w') as f:
            f.write(yml1)
        _, fout = tempfile.mkstemp()
        res = subprocess.run(['wcxf', 'convert', 'json', fin, '--output', fout])
        with open(fout, 'r') as f:
            d_json3 = json.load(f)
        self.assertDictEqual(d_yml1, d_json3)
        # delete temp files
        _del_files([fin, fout])
        # JSON file -> YAML file
        _, fin = tempfile.mkstemp()
        with open(fin, 'w') as f:
            f.write(json1)
        _, fout = tempfile.mkstemp()
        res = subprocess.run(['wcxf', 'convert', 'yaml', fin, '--output', fout])
        with open(fout, 'r') as f:
            yml3 = f.read()
        d_yml3 = yaml.load(yml3)
        self.assertDictEqual(d_yml3, d_json1)
        # delete temp files
        _del_files([fin, fout])

    def test_validate(self):
        _root = os.path.abspath(os.path.dirname(__file__))
        wet = os.path.join(_root, 'bases', 'wet.eft.json')
        with open(wet, 'rb') as f:
            s = f.read()
        res = subprocess.run(['wcxf', 'validate', 'eft', '-'],
                             input=s,
                             stdout=subprocess.PIPE)
        res = res.stdout.decode('utf-8')
        self.assertEqual(res, "Validation successful.\n")

        _root = os.path.abspath(os.path.dirname(__file__))
        wet = os.path.join(_root, 'bases', 'wet.flavio.basis.json')
        with open(wet, 'rb') as f:
            s = f.read()
        res = subprocess.run(['wcxf', 'validate', 'basis', '-'],
                             input=s,
                             stdout=subprocess.PIPE)
        res = res.stdout.decode('utf-8')
        self.assertEqual(res, "Validation successful.\n")



import numpy as np
from collections import OrderedDict, defaultdict
import pylha
import json
import yaml
from math import sqrt
import ckmutil
from wilson.util import smeftutil
from . import dsixtools_definitions as definitions
import wilson


def load(stream, fmt='lha'):
    """Load a parameter file in DSixTools SLHA-like format or its JSON or
    YAML representation."""
    if fmt == 'lha':
        return pylha.load(stream)
    elif fmt == 'json':
        if isinstance(stream, str):
            return json.loads(stream)
        else:
            return json.load(stream)
    elif fmt == 'yaml':
        return yaml.load(stream)

def lha2matrix(values, shape):
    """Return a matrix given a list of values of the form
    [[1, 1, float], [1, 2, float], ...]
    referring to the (1,1)-element etc.
    `shape` is the shape of the final matrix. All elements not provided
    will be assumed to be zero. Also works for higher-rank tensors."""
    M = np.zeros(shape)
    for v in values:
        M[tuple([int(i-1) for i in v[:-1]])] = v[-1]
    return M

def matrix2lha(M):
    """Inverse function to lha2matrix: return a LHA-like list given a tensor."""
    l = []
    ind = np.indices(M.shape).reshape(M.ndim, M.size).T
    for i in ind:
        l.append([j+1 for j in i] + [M[tuple(i)]])
    return l

def sm_lha2dict(lha):
    """Convert a dictionary returned by pylha from a DSixTools SM input file
    into a dictionary of SM values."""
    d = OrderedDict()
    v = dict(lha['BLOCK']['GAUGE']['values'])
    d['g'] = v[1]
    d['gp'] = v[2]
    d['gs'] = v[3]
    v = dict(lha['BLOCK']['SCALAR']['values'])
    d['Lambda'] = v[1]
    d['m2'] = v[2]
    d['Gu'] = lha2matrix(lha['BLOCK']['GU']['values'], (3,3))
    if 'IMGU' in lha['BLOCK']:
        d['Gu'] = d['Gu'] + 1j*lha2matrix(lha['BLOCK']['IMGU']['values'], (3,3))
    d['Gd'] = lha2matrix(lha['BLOCK']['GD']['values'], (3,3))
    if 'IMGD' in lha['BLOCK']:
        d['Gd'] = d['Gd'] + 1j*lha2matrix(lha['BLOCK']['IMGD']['values'], (3,3))
    d['Ge'] = lha2matrix(lha['BLOCK']['GE']['values'], (3,3))
    if 'IMGE' in lha['BLOCK']:
        d['Ge'] = d['Ge'] + 1j*lha2matrix(lha['BLOCK']['IMGE']['values'], (3,3))
    # thetas default to 0
    if 'THETA' in lha['BLOCK']:
        v = dict(lha['BLOCK']['THETA']['values'])
        d['Theta'] = v.get(1, 0)
        d['Thetap'] = v.get(2, 0)
        d['Thetas'] = v.get(3, 0)
    else:
        d['Theta'] = 0
        d['Thetap'] = 0
        d['Thetas'] = 0
    return d

def sm_dict2lha(d):
    """Convert a a dictionary of SM parameters into
    a dictionary that pylha can convert into a DSixTools SM output file."""
    blocks = OrderedDict([
        ('GAUGE', {'values': [[1, d['g'].real], [2, d['gp'].real], [3, d['gs'].real]]}),
        ('SCALAR', {'values': [[1, d['Lambda'].real], [2, d['m2'].real]]}),
        ('GU', {'values': matrix2lha(d['Gu'].real)}),
        ('IMGU', {'values': matrix2lha(d['Gu'].imag)}),
        ('GD', {'values': matrix2lha(d['Gd'].real)}),
        ('IMGD', {'values': matrix2lha(d['Gd'].imag)}),
        ('GE', {'values': matrix2lha(d['Ge'].real)}),
        ('IMGE', {'values': matrix2lha(d['Ge'].imag)}),
        ('THETA', {'values': [[1, d['Theta'].real], [2, d['Thetap'].real], [3, d['Thetas'].real]]}),
        ])
    return {'BLOCK': blocks}

# dictionary necessary for translating to DSixTools IO format
WC_dict_0f = OrderedDict([
('G', ('WC1', 1)),
('Gtilde', ('WC1', 2)),
('W', ('WC1', 3)),
('Wtilde', ('WC1', 4)),
('phi', ('WC2', 1)),
('phiBox', ('WC3', 1)),
('phiD', ('WC3', 2)),
('phiG', ('WC4', 1)),
('phiB', ('WC4', 2)),
('phiW', ('WC4', 3)),
('phiWB', ('WC4', 4)),
('phiGtilde', ('WC4', 5)),
('phiBtilde', ('WC4', 6)),
('phiWtilde', ('WC4', 7)),
('phiWtildeB', ('WC4', 8)),
])

def lha2scale(lha):
    """Extract the high scale from a dictionary eturned by pylha from
    a DSixTools options file."""
    return dict(lha['BLOCK']['SCALES']['values'])[1]

def wc_lha2dict(lha):
    """Convert a dictionary returned by pylha from a DSixTools WC input file
    into a dictionary of Wilson coefficients."""
    C = OrderedDict()
    # try to read all WCs with 0, 2, or 4 fermions; if not found, set to zero
    for k, (block, i) in WC_dict_0f.items():
        try:
            C[k] = dict(lha['BLOCK'][block]['values'])[i]
        except KeyError:
            C[k] = 0
    for k in smeftutil.WC_keys_2f:
        try:
            C[k] = lha2matrix(lha['BLOCK']['WC' + k.upper()]['values'], (3,3)).real
        except KeyError:
            C[k] = np.zeros((3,3))
        try: # try to add imaginary part
            C[k] = C[k] + 1j*lha2matrix(lha['BLOCK']['IMWC' + k.upper()]['values'], (3,3))
        except KeyError:
            pass
    for k in smeftutil.WC_keys_4f:
        try:
            C[k] = lha2matrix(lha['BLOCK']['WC' + k.upper()]['values'], (3,3,3,3))
        except KeyError:
            C[k] = np.zeros((3,3,3,3))
        try: # try to add imaginary part
            C[k] = C[k] + 1j*lha2matrix(lha['BLOCK']['IMWC' + k.upper()]['values'], (3,3,3,3))
        except KeyError:
            pass
    return C

def wc_dict2lha(wc, skip_redundant=True, skip_zero=True):
    """Convert a a dictionary of Wilson coefficients into
    a dictionary that pylha can convert into a DSixTools WC output file."""
    d = OrderedDict()
    for name, (block, i) in WC_dict_0f.items():
        if block not in d:
            d[block] = defaultdict(list)
        if wc[name] != 0:
            d[block]['values'].append([i, wc[name].real])
    for name in smeftutil.WC_keys_2f:
        reblock = 'WC'+name.upper()
        imblock = 'IMWC'+name.upper()
        if reblock not in d:
            d[reblock] = defaultdict(list)
        if imblock not in d:
            d[imblock] = defaultdict(list)
        for i in range(3):
            for j in range(3):
                if (i, j) in definitions.redundant_elements[name] and skip_redundant:
                    # skip redundant elements
                    continue
                if wc[name][i, j].real != 0 or not skip_zero:
                    d[reblock]['values'].append([i+1, j+1, float(wc[name][i, j].real)])
                if wc[name][i, j].imag != 0 or not skip_zero:
                    # omit Im parts that have to vanish by symmetry
                    if (i, j) not in definitions.vanishing_im_parts[name]:
                        d[imblock]['values'].append([i+1, j+1, float(wc[name][i, j].imag)])
    for name in smeftutil.WC_keys_4f:
        reblock = 'WC'+name.upper()
        imblock = 'IMWC'+name.upper()
        if reblock not in d:
            d[reblock] = defaultdict(list)
        if imblock not in d:
            d[imblock] = defaultdict(list)
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    for l in range(3):
                        if (i, j, k, l) in definitions.redundant_elements[name] and skip_redundant:
                            # skip redundant elements
                            continue
                        if wc[name][i, j, k, l].real != 0 or not skip_zero:
                            d[reblock]['values'].append([i+1, j+1, k+1, l+1, float(wc[name][i, j, k, l].real)])
                        if wc[name][i, j, k, l].imag != 0 or not skip_zero:
                            # omit Im parts that have to vanish by symmetry
                            if (i, j, k, l) not in definitions.vanishing_im_parts[name]:
                                d[imblock]['values'].append([i+1, j+1, k+1, l+1, float(wc[name][i, j, k, l].imag)])
    # remove empty blocks
    empty = []
    for block in d:
        if d[block] == {}:
            empty.append(block)
    for block in empty:
        del d[block]
    return {'BLOCK': d}


class SMEFTio(object):

    def __init__(self):
        """Initialize the SMEFTio instance."""
        self.C_in = None
        self.scale_in = None
        self.scale_high = None

    def set_initial(self, C_in, scale_in, scale_high):
        r"""Set the initial values for parameters and Wilson coefficients at
        the scale `scale_in`, setting the new physics scale $\Lambda$ to
        `scale_high`."""
        self.C_in = C_in
        self.scale_in = scale_in
        self.scale_high = scale_high

    def load_initial(self, streams):
        """Load the initial values for parameters and Wilson coefficients from
        one or several files.

        `streams` should be a tuple of file-like objects strings."""
        d = {}
        for stream in streams:
            s = load(stream)
            if 'BLOCK' not in s:
                raise ValueError("No BLOCK found")
            d.update(s['BLOCK'])
        d = {'BLOCK': d}
        C = wc_lha2dict(d)
        sm = sm_lha2dict(d)
        self.scale_high = lha2scale(d)
        self.scale_in = lha2scale(d)
        C.update(sm)
        C = smeftutil.symmetrize(C)
        self.C_in = C

    def set_initial_wcxf(self, wc, scale_high=None):
        """Load the initial values for Wilson coefficients from a
        wcxf.WC instance.

        Parameters:

        - `scale_high`: since Wilson coefficients are dimensionless in
          DsixTools but not in WCxf, the high scale in GeV has to be provided.
          If this parameter is None (default), either a previously defined
          value will be used, or the scale attribute of the WC instance will
          be used.
        """
        import wcxf
        if wc.eft != 'SMEFT':
            raise ValueError("Wilson coefficients use wrong EFT.")
        if wc.basis != 'Warsaw':
            raise ValueError("Wilson coefficients use wrong basis.")
        if scale_high is not None:
            self.scale_high = scale_high
        elif self.scale_high is None:
            self.scale_high = wc.scale
        C = wilson.translate.smeft.wcxf2arrays(wc.dict)
        keys_dim5 = ['llphiphi']
        keys_dim6 = list(set(smeftutil.WC_keys_0f + smeftutil.WC_keys_2f + smeftutil.WC_keys_4f) - set(keys_dim5))
        self.scale_in = wc.scale
        for k in keys_dim5:
            if k in C:
                C[k] = C[k]*self.scale_high
        for k in keys_dim6:
            if k in C:
                C[k] = C[k]*self.scale_high**2
        C = smeftutil.symmetrize(C)
        # fill in zeros for missing WCs
        for k, s in smeftutil.C_keys_shape.items():
            if k not in C and k not in smeftutil.SM_keys:
                if s == 1:
                    C[k] = 0
                else:
                    C[k] = np.zeros(s)
        if self.C_in is None:
            self.C_in = C
        else:
            self.C_in.update(C)

    def load_wcxf(self, stream):
        """Load the initial values for Wilson coefficients from
        a file-like object or a string in WCxf format.

        Note that Standard Model parameters have to be provided separately
        and are assumed to be in the weak basis used for the Warsaw basis as
        defined in WCxf, i.e. in the basis where the down-type and charged
        lepton mass matrices are diagonal."""
        import wcxf
        wc = wcxf.WC.load(stream)
        self.set_initial_wcxf(wc)

    def dump(self, C_out, scale_out=None, stream=None, fmt='lha', skip_redundant=True):
        """Return a string representation of the parameters and Wilson
        coefficients `C_out` in DSixTools output format. If `stream` is
        specified, export it to a file. `fmt` defaults to `lha` (the SLHA-like
        DSixTools format), but can also be `json` or `yaml` (see the
        pylha documentation)."""
        C = OrderedDict()
        if scale_out is not None:
            C['SCALES'] = {'values': [[1, self.scale_high], [2, scale_out]]}
        else:
            C['SCALES'] = {'values': [[1, self.scale_high]]}
        # sm = sm_dict2lha(C_out)['BLOCK']
        # C.update(sm)
        wc = wc_dict2lha(C_out, skip_redundant=skip_redundant)['BLOCK']
        C.update(wc)
        return pylha.dump({'BLOCK': C}, fmt=fmt, stream=stream)

    def get_wcxf(self, C_out, scale_out):
        """Return the Wilson coefficients `C_out` as a wcxf.WC instance.

        Note that the Wilson coefficients are rotated into the Warsaw basis
        as defined in WCxf, i.e. to the basis where the down-type and charged
        lepton mass matrices are diagonal."""
        import wcxf
        # C = self.rotate_defaultbasis(C_out)
        C = C_out.copy()  # FIXME
        d = wilson.translate.smeft.arrays2wcxf(C)
        basis = wcxf.Basis['SMEFT', 'Warsaw']
        d = {k: v for k, v in d.items() if k in basis.all_wcs and v != 0}
        keys_dim5 = ['llphiphi']
        keys_dim6 = list(set(smeftutil.WC_keys_0f + smeftutil.WC_keys_2f
                             + smeftutil.WC_keys_4f) - set(keys_dim5))
        for k in d:
            if k.split('_')[0] in keys_dim5:
                d[k] = d[k] / self.scale_high
        for k in d:
            if k.split('_')[0] in keys_dim6:
                d[k] = d[k] / self.scale_high**2
        # d = {k: v for k, v in d.items() if v != 0}
        d = wcxf.WC.dict2values(d)
        wc = wcxf.WC('SMEFT', 'Warsaw', scale_out, d)
        return wc

    def dump_wcxf(self, C_out, scale_out, fmt='yaml', stream=None, **kwargs):
        """Return a string representation of the Wilson coefficients `C_out`
        in WCxf format. If `stream` is specified, export it to a file.
        `fmt` defaults to `yaml`, but can also be `json`.

        Note that the Wilson coefficients are rotated into the Warsaw basis
        as defined in WCxf, i.e. to the basis where the down-type and charged
        lepton mass matrices are diagonal."""
        wc = self.get_wcxf(C_out, scale_out)
        return wc.dump(fmt=fmt, stream=stream, **kwargs)

    def rotate_defaultbasis(self, C):
        """Rotate all parameters to the basis where the running down-type quark
        and charged lepton mass matrices are diagonal and where the running
        up-type quark mass matrix has the form V.S, with V unitary and S real
        diagonal, and where the CKM and PMNS matrices have the standard
        phase convention."""
        v = 246.22
        Mep = v/sqrt(2) * (C['Ge'] - C['ephi'] * v**2/self.scale_high**2/2)
        Mup = v/sqrt(2) * (C['Gu'] - C['uphi'] * v**2/self.scale_high**2/2)
        Mdp = v/sqrt(2) * (C['Gd'] - C['dphi'] * v**2/self.scale_high**2/2)
        Mnup = -v**2 * C['llphiphi']
        UeL, Me, UeR = ckmutil.diag.msvd(Mep)
        UuL, Mu, UuR = ckmutil.diag.msvd(Mup)
        UdL, Md, UdR = ckmutil.diag.msvd(Mdp)
        Unu, Mnu = ckmutil.diag.mtakfac(Mnup)
        UuL, UdL, UuR, UdR = ckmutil.phases.rephase_standard(UuL, UdL, UuR, UdR)
        Unu, UeL, UeR = ckmutil.phases.rephase_pmns_standard(Unu, UeL, UeR)
        return smeftutil.flavor_rotation(C, Uq=UdL, Uu=UuR, Ud=UdR, Ul=UeL, Ue=UeR)


def wcxf2dsixtools(wc, stream=None):
    smeftio = SMEFTio()
    smeftio.set_initial_wcxf(wc)
    return smeftio.dump(smeftio.C_in, stream=stream)

def dsixtools2wcxf(streams, stream=None):
    smeftio = SMEFTio()
    smeftio.load_initial(streams)
    return smeftio.dump_wcxf(smeftio.C_in, smeftio.scale_in, stream=stream)



import unittest
import wcxf
import pkgutil
class TestWET(unittest.TestCase):
    def test_bern(self):
        basis = wcxf.Basis['WET', 'Bern']
        # comparing no. of WCs to table 1 of arXiv:1704.06639
        self.assertEqual(len(basis.sectors['sbsb']), 8)
        self.assertEqual(len(basis.sectors['dbdb']), 8)
        self.assertEqual(len(basis.sectors['ubenu']), 5*3)
        self.assertEqual(len(basis.sectors['ubmunu']), 5*3)
        self.assertEqual(len(basis.sectors['ubtaunu']), 5*3)
        self.assertEqual(len(basis.sectors['cbenu']), 5*3)
        self.assertEqual(len(basis.sectors['cbmunu']), 5*3)
        self.assertEqual(len(basis.sectors['cbtaunu']), 5*3)
        self.assertEqual(len(basis.sectors['sbuc']), 20)
        self.assertEqual(len(basis.sectors['sbcu']), 20)
        self.assertEqual(len(basis.sectors['dbuc']), 20)
        self.assertEqual(len(basis.sectors['dbcu']), 20)
        self.assertEqual(len(basis.sectors['sbsd']), 10)
        self.assertEqual(len(basis.sectors['dbds']), 10)
        self.assertEqual(len(basis.sectors['sb']), 57*2)
        self.assertEqual(len(basis.sectors['sbnunu']), 2*9)
        self.assertEqual(len(basis.sectors['db']), 57*2)
        self.assertEqual(len(basis.sectors['dbnunu']), 2*9)
        for l1 in ['e', 'mu', 'tau']:
            for l2 in ['e', 'mu', 'tau']:
                if l1 != l2:
                    self.assertEqual(len(basis.sectors['sb'+l1+l2]), 10)
                    self.assertEqual(len(basis.sectors['db'+l1+l2]), 10)

    def test_jms(self):
        basis = wcxf.Basis['WET', 'JMS']
        all_wc = {k: v for sk, sv in basis.sectors.items() for k, v in sv.items()}
        def isreal(wc):
            if 'real' in all_wc[wc] and all_wc[wc]['real']:
                return True
            else:
                return False
        def assert_len(s, n):
            L = sum([1 if isreal(k) else 2 for k in all_wc if s in k])
            try:
                self.assertEqual(L, n,
                             msg="Failed for {}".format(s))
            except Exception as exc:
                print(exc)
        # compare individual counts to tables 11-17 in arXiv:1709.04486
        assert_len('egamma_', 9*2)
        assert_len('ugamma_', 4*2)
        assert_len('dgamma_', 9*2)
        assert_len('uG_', 4*2)
        assert_len('dG_', 9*2)
        assert_len('VnunuLL_', 36)
        assert_len('VeeLL_', 36)
        assert_len('VnueLL_', 81)
        assert_len('VnuuLL_', 36)
        assert_len('VnudLL_', 81)
        assert_len('VeuLL_', 36)
        assert_len('VedLL_', 81)
        assert_len('VnueduLL_', 2*54)
        assert_len('VuuLL_', 10)
        assert_len('VddLL_', 45)
        assert_len('V1udLL_', 36)
        assert_len('V8udLL_', 36)
        assert_len('VeeRR_', 36)
        assert_len('VeuRR_', 36)
        assert_len('VedRR_', 81)
        assert_len('VuuRR_', 10)
        assert_len('VddRR_', 45)
        assert_len('V1udRR_', 36)
        assert_len('V8udRR_', 36)
        assert_len('VnueLR_', 81)
        assert_len('VeeLR_', 81)
        assert_len('VnuuLR_', 36)
        assert_len('VnudLR_', 81)
        assert_len('VeuLR_', 36)
        assert_len('VedLR_', 81)
        assert_len('VueLR_', 36)
        assert_len('VdeLR_', 81)
        assert_len('VnueduLR_', 2*54)
        assert_len('V1uuLR_', 16)
        assert_len('V8uuLR_', 16)
        assert_len('V1udLR_', 36)
        assert_len('V8udLR_', 36)
        assert_len('V1duLR_', 36)
        assert_len('V8duLR_', 36)
        assert_len('V1ddLR_', 81)
        assert_len('V8ddLR_', 81)
        assert_len('V1udduLR_', 2*36)
        assert_len('V8udduLR_', 2*36)
        assert_len('SeuRL_', 2*36)
        assert_len('SedRL_', 2*81)
        assert_len('SnueduRL_', 2*54)
        assert_len('SeeRR', 2*45)
        assert_len('SeuRR', 2*36)
        assert_len('TeuRR', 2*36)
        assert_len('SedRR', 2*81)
        assert_len('TedRR', 2*81)
        assert_len('SnueduRR', 2*54)
        assert_len('TnueduRR', 2*54)
        assert_len('S1uuRR', 2*10)
        assert_len('S8uuRR', 2*10)
        assert_len('S1udRR', 2*36)
        assert_len('S8udRR', 2*36)
        assert_len('S1ddRR', 2*45)
        assert_len('S8ddRR', 2*45)
        assert_len('S1udduRR', 2*36)
        assert_len('S8udduRR', 2*36)
        # compare total count to table 22
        ntot = sum([1 if isreal(k) else 2 for k in all_wc])
        self.assertEqual(ntot, 1+87+186+76+21+66+76+90+252+266+171+45+342+254+1+66+156+51+15+51+51+72+207+216+171+45+342+254+9+9+26+26)

#   Unit test for Bern to flavio translator
    def test_wc1(self):
        f = pkgutil.get_data('wcxf', 'data/test.wcs.bern.yml').decode('utf-8')
        wc0 = wcxf.WC.load(f)
        wc0.validate()
        self.assertEqual(wc0.eft, 'WET')
        self.assertEqual(wc0.basis, 'Bern')
        self.assertEqual(wc0.scale, 4.8)
        wc1= wc0.translate('flavio')
        f = pkgutil.get_data('wcxf', 'data/test.wcs.flavio.out.yml').decode('utf-8')
        wc2 = wcxf.WC.load(f)
        self.assertEqual(wc2.eft, 'WET')
        self.assertEqual(wc2.basis, 'flavio')
        self.assertEqual(wc2.scale, 4.8)
        wc2.validate()
        self.assertAlmostEqual(wc1.dict['CVLL_sdsd'], wc2.dict['CVLL_sdsd'])
        self.assertAlmostEqual(wc1.dict['CVLL_bsbs'], wc2.dict['CVLL_bsbs'])
        self.assertAlmostEqual(wc1.dict['CVLL_bdbd'], wc2.dict['CVLL_bdbd'])
        self.assertAlmostEqual(wc1.dict['CVL_dumunumu'], wc2.dict['CVL_dumunumu'])
        self.assertAlmostEqual(wc1.dict['CVL_sumunumu'], wc2.dict['CVL_sumunumu'])
        self.assertAlmostEqual(wc1.dict['CVL_bumunumu'], wc2.dict['CVL_bumunumu'])
        self.assertAlmostEqual(wc1.dict['CVL_dcmunumu'], wc2.dict['CVL_dcmunumu'])
        self.assertAlmostEqual(wc1.dict['CVL_scmunumu'], wc2.dict['CVL_scmunumu'])
        self.assertAlmostEqual(wc1.dict['CVL_bcmunumu'], wc2.dict['CVL_bcmunumu'])
        self.assertAlmostEqual(wc1.dict['C9_bsmumu'], wc2.dict['C9_bsmumu'])
        self.assertAlmostEqual(wc1.dict['C9_bdmumu'], wc2.dict['C9_bdmumu'])
        self.assertAlmostEqual(wc1.dict['C7_bs'], wc2.dict['C7_bs'])
        self.assertAlmostEqual(wc1.dict['C8_bs'], wc2.dict['C8_bs'])
        self.assertAlmostEqual(wc1.dict['C7_bd'], wc2.dict['C7_bd'])
        self.assertAlmostEqual(wc1.dict['C8_bd'], wc2.dict['C8_bd'])

#   Unit test for flavio to Bern translator
    def test_wc2(self):
        f = pkgutil.get_data('wcxf', 'data/test.wcs.flavio.yml').decode('utf-8')
        wc0 = wcxf.WC.load(f)
        wc0.validate()
        self.assertEqual(wc0.eft, 'WET')
        self.assertEqual(wc0.basis, 'flavio')
        self.assertEqual(wc0.scale, 4.8)
        wc1= wc0.translate('Bern')
        f = pkgutil.get_data('wcxf', 'data/test.wcs.bern.out.yml').decode('utf-8')
        wc2 = wcxf.WC.load(f)
        self.assertEqual(wc2.eft, 'WET')
        self.assertEqual(wc2.basis, 'Bern')
        self.assertEqual(wc2.scale, 4.8)
        wc2.validate()
        self.assertAlmostEqual(wc1.dict['1sbsb'], wc2.dict['1sbsb'])
        self.assertAlmostEqual(wc1.dict['1dbdb'], wc2.dict['1dbdb'])
        self.assertAlmostEqual(wc1.dict['1dsds'], wc2.dict['1dsds'])
        self.assertAlmostEqual(wc1.dict['1udmumu'], wc2.dict['1udmumu'])
        self.assertAlmostEqual(wc1.dict['1usmumu'], wc2.dict['1usmumu'])
        self.assertAlmostEqual(wc1.dict['1ubmumu'], wc2.dict['1ubmumu'])
        self.assertAlmostEqual(wc1.dict['1cdmumu'], wc2.dict['1cdmumu'])
        self.assertAlmostEqual(wc1.dict['1csmumu'], wc2.dict['1csmumu'])
        self.assertAlmostEqual(wc1.dict['1cbmumu'], wc2.dict['1cbmumu'])
        self.assertAlmostEqual(wc1.dict['1sbmumu'], wc2.dict['1sbmumu'])
        self.assertAlmostEqual(wc1.dict['1dbmumu'], wc2.dict['1dbmumu'])
        self.assertAlmostEqual(wc1.dict['7gammasb'], wc2.dict['7gammasb'])
        self.assertAlmostEqual(wc1.dict['8gsb'], wc2.dict['8gsb'])
        self.assertAlmostEqual(wc1.dict['7gammadb'], wc2.dict['7gammadb'])
        self.assertAlmostEqual(wc1.dict['8gdb'], wc2.dict['8gdb'])
        
#   Unit test for JMS to Bern translator
    def test_wc3(self):
        f = pkgutil.get_data('wcxf', 'data/test.wcs.jms.yml').decode('utf-8')
        wc0 = wcxf.WC.load(f)
        wc0.validate()
        self.assertEqual(wc0.eft, 'WET')
        self.assertEqual(wc0.basis, 'JMS')
        self.assertEqual(wc0.scale, 4.8)
        wc1= wc0.translate('Bern')
        f = pkgutil.get_data('wcxf', 'data/test.wcs.bern2.out.yml').decode('utf-8')
        wc2 = wcxf.WC.load(f)
        self.assertEqual(wc2.eft, 'WET')
        self.assertEqual(wc2.basis, 'Bern')
        self.assertEqual(wc2.scale, 4.8)
        wc2.validate()
        self.assertAlmostEqual(wc1.dict['1sbsb'], wc2.dict['1sbsb'])
        self.assertAlmostEqual(wc1.dict['1dbdb'], wc2.dict['1dbdb'])
        self.assertAlmostEqual(wc1.dict['1dsds'], wc2.dict['1dsds'])
        self.assertAlmostEqual(wc1.dict['1udmumu'], wc2.dict['1udmumu'])
        self.assertAlmostEqual(wc1.dict['1usmumu'], wc2.dict['1usmumu'])
        self.assertAlmostEqual(wc1.dict['1ubmumu'], wc2.dict['1ubmumu'])
        self.assertAlmostEqual(wc1.dict['1cdmumu'], wc2.dict['1cdmumu'])
        self.assertAlmostEqual(wc1.dict['1csmumu'], wc2.dict['1csmumu'])
        self.assertAlmostEqual(wc1.dict['1cbmumu'], wc2.dict['1cbmumu'])
        self.assertAlmostEqual(wc1.dict['1sbuc'], wc2.dict['1sbuc'])
        self.assertAlmostEqual(wc1.dict['3sbuc'], wc2.dict['3sbuc'])
        self.assertAlmostEqual(wc1.dict['1sbcu'], wc2.dict['1sbcu'])
        self.assertAlmostEqual(wc1.dict['3sbcu'], wc2.dict['3sbcu'])
        self.assertAlmostEqual(wc1.dict['1dbuc'], wc2.dict['1dbuc'])
        self.assertAlmostEqual(wc1.dict['3dbuc'], wc2.dict['3dbuc'])
        self.assertAlmostEqual(wc1.dict['1dbcu'], wc2.dict['1dbcu'])
        self.assertAlmostEqual(wc1.dict['3dbcu'], wc2.dict['3dbcu'])
        self.assertAlmostEqual(wc1.dict['1dsuc'], wc2.dict['1dsuc'])
        self.assertAlmostEqual(wc1.dict['3dsuc'], wc2.dict['3dsuc'])
        self.assertAlmostEqual(wc1.dict['1dscu'], wc2.dict['1dscu'])
        self.assertAlmostEqual(wc1.dict['3dscu'], wc2.dict['3dscu'])
        self.assertAlmostEqual(wc1.dict['1sbsd'], wc2.dict['1sbsd'])
        self.assertAlmostEqual(wc1.dict['3sbsd'], wc2.dict['3sbsd'])
        self.assertAlmostEqual(wc1.dict['1dbds'], wc2.dict['1dbds'])
        self.assertAlmostEqual(wc1.dict['3dbds'], wc2.dict['3dbds'])
#        self.assertAlmostEqual(wc1.dict['1dbsb'], wc2.dict['1dbsb'])
#        self.assertAlmostEqual(wc1.dict['3dbsb'], wc2.dict['3dbsb'])
        self.assertAlmostEqual(wc1.dict['1sbmumu'], wc2.dict['1sbmumu'])
        self.assertAlmostEqual(wc1.dict['3sbmumu'], wc2.dict['3sbmumu'])
        self.assertAlmostEqual(wc1.dict['1dbmumu'], wc2.dict['1dbmumu'])
        self.assertAlmostEqual(wc1.dict['3dbmumu'], wc2.dict['3dbmumu'])
        self.assertAlmostEqual(wc1.dict['1dsmumu'], wc2.dict['1dsmumu'])
        self.assertAlmostEqual(wc1.dict['3dsmumu'], wc2.dict['3dsmumu'])
        self.assertAlmostEqual(wc1.dict['1sbdd'], wc2.dict['1sbdd'])
        self.assertAlmostEqual(wc1.dict['3sbdd'], wc2.dict['3sbdd'])
        self.assertAlmostEqual(wc1.dict['1dbdd'], wc2.dict['1dbdd'])
        self.assertAlmostEqual(wc1.dict['3dbdd'], wc2.dict['3dbdd'])
        self.assertAlmostEqual(wc1.dict['1dsdd'], wc2.dict['1dsdd'])
        self.assertAlmostEqual(wc1.dict['3dsdd'], wc2.dict['3dsdd'])
        self.assertAlmostEqual(wc1.dict['7gammasb'], wc2.dict['7gammasb'])
        self.assertAlmostEqual(wc1.dict['8gsb'], wc2.dict['8gsb'])



import wcxf
import yaml
import json
import logging


def convert_json(stream_in, stream_out):
    try:
        return wcxf.classes._yaml_to_json(stream_in, stream_out, indent=2)
    except yaml.YAMLError:
        logging.error("Input file cannot be parsed as YAML.")
        return 1


def convert_yaml(stream_in, stream_out):
    try:
        return wcxf.classes._json_to_yaml(stream_in, stream_out,
                                          default_flow_style=False)
    except json.decoder.JSONDecodeError:
        logging.error("Input file cannot be parsed as JSON.")
        return 1



"""Functions to convert WCxf files to EOS parameter files."""

from collections import OrderedDict
import yaml
import glob
import os
import re


def get_sm_wcs(eos_parameter_dir):
    """Read SM contributions to EOS Wilson coefficients from the EOS
    installation parameter directory and return them as a dictionary."""
    all_wcs = {}
    yamlfiles = glob.glob(os.path.join(eos_parameter_dir, '*.yaml'))
    for yamlfile in yamlfiles:
        with open(yamlfile, 'r') as f:
            wcs = yaml.load(f)
        meta = wcs.get('@metadata@', {})
        if 'wcxf-relevant' in meta and meta['wcxf-relevant']:
            del wcs['@metadata@']
            all_wcs.update(wcs)
    return all_wcs


def get_eos_np(sm_key, wc):
    """Get the NP contribution to an EOS Wilson coefficient identified
    by the name `sm_key` from a wcxf.WC instance `wc`"""
    if sm_key in wc.dict:
        return wc.dict[sm_key].real
    elif r'Re{' in sm_key: # if the EOS name contains "Re"
        k2 = re.sub(r'Re\{([^\}]+)\}', r'\1', sm_key) # name of the coeff in the basis file
        if k2 in wc.dict:
            return wc.dict[k2].real
        elif k2.replace('lnu', 'enue') in wc.dict or k2.replace('lnu', 'munumu') in wc.dict:
            # for charged-current WCs, check if the WCxf file is LFU, else raise an error
            if wc.dict.get(k2.replace('lnu', 'enue'), 0) != wc.dict.get(k2.replace('lnu', 'munumu'), 0):
                raise ValueError("Found lepton flavour non-universal charged-current coefficients")
            return wc.dict.get(k2.replace('lnu', 'enue'), 0).real
    elif r'Im{' in sm_key: # if the EOS name contains "Im"
        k2 = re.sub(r'Im\{([^\}]+)\}', r'\1', sm_key) # name of the coeff in the basis file
        if k2 in wc.dict:
            return wc.dict[k2].imag
        elif k2.replace('lnu', 'enue') in wc.dict or k2.replace('lnu', 'munumu') in wc.dict:
            # for charged-current WCs, check if the WCxf file is LFU, else raise an error
            if wc.dict.get(k2.replace('lnu', 'enue'), 0) != wc.dict.get(k2.replace('lnu', 'munumu'), 0):
                raise ValueError("Found lepton flavour non-universal charged-current coefficients")
            return wc.dict.get(k2.replace('lnu', 'enue'), 0).imag
    return 0


def wcxf2eos(wc, sm_wc_dict):
    """From a wcxf.WC instance wc and a dictionary of EOS Wilson coefficient
    SM contributions, return a dictionary of EOS Wilson coefficient parameter
    values including SM and NP contributions."""
    eosd = OrderedDict()
    for k, v in sm_wc_dict.items():
        sm = v['central']
        np = get_eos_np(k, wc)
        eosd[k] = OrderedDict()
        for q in ['central', 'min', 'max']:
            eosd[k][q] = sm + np
    return eosd



import unittest
import numpy as np
import numpy.testing as npt
import wcxf
from wcxf.converters import eos
import os

my_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(my_path, '..', 'data')

class TestEOS(unittest.TestCase):
    def test_get_sm_wcs(self):
        eos_sm_dict = eos.get_sm_wcs(data_path)
        self.assertEqual(eos_sm_dict['b->s::Re{c7}']['central'], -0.33726473)
        self.assertEqual(eos_sm_dict['b->s::c5']['central'], 0.00042854)
        self.assertEqual(eos_sm_dict['b->ulnu::Im{cVL}']['min'], 0.0)
    
    def test_wcxf2eos(self):
        sm_wc_dict = eos.get_sm_wcs(data_path)
        wc_dict = { 'b->s::c7': 0.1,
                    'b->uenue::cSL': 0.2j, 'b->umunumu::cSL': 0.2j,
                    'b->uenue::cVL': 0.2, 'b->umunumu::cVL': 0.2,
                    'b->s::c5': -0.03 }
        wc = wcxf.WC('WET', 'EOS', 4.2, wcxf.WC.dict2values(wc_dict))
        eos_dict = eos.wcxf2eos(wc, sm_wc_dict)
        self.assertEqual(eos_dict['b->s::Re{c7}']['central'], -0.33726473 + 0.1)
        self.assertEqual(eos_dict['b->s::c5']['central'], 0.00042854 - 0.03)
        self.assertEqual(eos_dict['b->ulnu::Im{cSL}']['min'], 0.0 + 0.2)
        self.assertEqual(eos_dict['b->ulnu::Re{cVL}']['max'], 1.0 + 0.2)
        
    def test_lfu(self):
        sm_wc_dict = eos.get_sm_wcs(data_path)
        wc_dict = { 'b->uenue::cSL': 0.2j, 'b->umunumu::cSL': 0.1j, }
        wc = wcxf.WC('WET', 'EOS', 4.2, wcxf.WC.dict2values(wc_dict))
        with self.assertRaises(ValueError):
            eos.wcxf2eos(wc, sm_wc_dict)


from setuptools import setup, find_packages

setup(name='wcxf',
      version='1.4.2',
      author='David M. Straub, Jason Aebischer',
      author_email='david.straub@tum.de, jason.aebischer@tum.de',
      license='MIT',
      packages=find_packages(),
      package_data={
        'wcxf': ['data/*.yml',
                 'data/*.yaml',
                 'data/*.json',
                 'bases/*.json',
                 'bases/child/*.json',
                ]
      },
      install_requires=['pyyaml', 'ckmutil>=0.3', 'rundec>=0.5', 'pandas',
                        'wilson'],
      extras_require={
            'testing': ['nose', 'smeftrunner'],
      },
      entry_points={
        'console_scripts': [
            'wcxf = wcxf.cli:wcxf_cli',
            'wcxf2eos = wcxf.cli:eos',
            'wcxf2dsixtools = wcxf.cli:wcxf2dsixtools',
            'dsixtools2wcxf = wcxf.cli:dsixtools2wcxf',
            'wcxf2smeftsim = wcxf.cli:smeftsim',
        ]
      },
    )



import argparse
import wcxf
import sys
import logging
import os
import yaml
import pylha


def wcxf_cli():
    parser = argparse.ArgumentParser(description="Command line interface to manipulate WCxf files.")
    subparsers = parser.add_subparsers(title='subcommands')

    # convert

    parser_convert = subparsers.add_parser('convert',
                                           description="Command line script to convert WCxf files between YAML and JSON.",
                                           help="convert between YAML and JSON formats")
    parser_convert.add_argument("FORMAT", type=str,
                                help="Output format (should be yaml or json)")
    parser_convert.add_argument("FILE", nargs='?', type=argparse.FileType('r'),
                                default=sys.stdin,
                                help="Input file. If \"-\", read from standard input")
    parser_convert.add_argument("--output", nargs='?',
                                      type=argparse.FileType('w'),
                                      default=sys.stdout,
                                      help="Output file. If absent, print to standard output")
    parser_convert.set_defaults(func=convert)

    # translate

    parser_translate = subparsers.add_parser('translate',
                                             description="Command line script for basis translation of WCxf files.",
                                             help="Translate between different bases")
    parser_translate.add_argument("BASIS", help="Output basis", type=str)
    parser_translate.add_argument("FILE", nargs='?',
                                  type=argparse.FileType('r'),
                                  default=sys.stdin,
                                  help="Input file. If \"-\", read from standard input")
    parser_translate.add_argument("--output", nargs='?',
                                  type=argparse.FileType('w'), default=sys.stdout,
                                  help="Output file. If absent, print to standard output")
    parser_translate.add_argument("--format", type=str,
                                  default="json",
                                  help="Output format (default: json)")
    parser_translate.set_defaults(func=translate)

    # match

    parser_match = subparsers.add_parser('match',
                                         description="Command line script for matching of WCxf files.",
                                         help="Match between different EFTs")
    parser_match.add_argument("EFT", help="Output EFT", type=str)
    parser_match.add_argument("BASIS", help="Output basis", type=str)
    parser_match.add_argument("FILE", nargs='?',
                              type=argparse.FileType('r'), default=sys.stdin,
                              help="Input file. If \"-\", read from standard input")
    parser_match.add_argument("--output", nargs='?',
                              type=argparse.FileType('w'), default=sys.stdout,
                              help="Output file. If absent, print to standard output")
    parser_match.add_argument("--format", type=str, default="json",
                              help="Output format (default: json)")
    parser_match.set_defaults(func=match)

    # validate

    parser_validate = subparsers.add_parser('validate',
                                            description="Command line script for validation of WCxf files.",
                                            help="Validate basis or Wilson coefficient files")
    parser_validate.add_argument("TYPE", type=str,
                                       help="Type of file to validate: should be 'eft', 'basis', or 'wc'")
    parser_validate.add_argument("FILE", nargs='?',
                                 type=argparse.FileType('r'), default=sys.stdin,
                                 help="Input file. If \"-\", read from standard input")
    parser_validate.set_defaults(func=validate)

    args = parser.parse_args()
    try:
        args.func(args)
    except AttributeError:
        parser.print_help()


def convert(args):
    from wcxf.converters.yamljson import convert_json, convert_yaml
    if args.FORMAT.lower() == 'json':
        convert_json(args.FILE, args.output)
    if args.FORMAT.lower() == 'yaml':
        convert_yaml(args.FILE, args.output)


def translate(args):
    wc_in = wcxf.WC.load(args.FILE)
    wc_out = wc_in.translate(args.BASIS)
    wc_out.dump(stream=args.output, fmt=args.format)


def match(args):
    wc_in = wcxf.WC.load(args.FILE)
    wc_out = wc_in.match(args.EFT, args.BASIS)
    wc_out.dump(stream=args.output, fmt=args.format)


def validate(args):
    if args.TYPE == 'eft':
        eft = wcxf.EFT.load(args.FILE)
    elif args.TYPE == 'basis':
        basis = wcxf.Basis.load(args.FILE)
        basis.validate()
    elif args.TYPE == 'wc':
        wc = wcxf.WC.load(args.FILE)
        wc.validate()
    else:
        logging.error("TYPE should be 'eft', 'basis', or 'wc'")
        return 1
    print("Validation successful.")
    return 0


def eos():
    from wcxf.converters.eos import wcxf2eos, get_sm_wcs
    parser = argparse.ArgumentParser(description="""Command line script to convert a WCxf file to an EOS Wilson coefficient parameter file.""",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("FILE", nargs='?', help="Input file. If \"-\", read from standard input",
                        type=argparse.FileType('r'), default=sys.stdin)
    parser.add_argument("--eosprefix", help="Installation prefix for the EOS installation. Defaults to /usr",
                        default='/usr')
    parser.add_argument("--output", nargs='?', help="Output file. If absent, print to standard output",
                        type=argparse.FileType('w'), default=sys.stdout)
    parser.add_argument("--eoshome", help="EOS home directory. If specified, values will be written to EOSHOME/parameters/wcxf.yaml. Cannot be used simultaneously with output",
                        default=None)
    args = parser.parse_args()
    # check sanity of inputs
    if args.output != sys.stdout and args.eoshome is not None:
        logging.error("Cannot use --output and --eoshome arguments simultaneously")
        return 1
    elif args.output == sys.stdout and args.eoshome is not None:
        output_dir = os.path.join(args.eoshome, 'parameters')
        if not os.path.isdir(output_dir):
            logging.error("Output directory {} does not exist".format(output_dir))
            return 1
        f = open(os.path.join(output_dir, 'wcxf.yaml'), 'w')
    else:
        f = args.output
    # read in & validate WCxf file
    wc = wcxf.WC.load(args.FILE)
    wc.validate()
    # read EOS SM contributions
    sm_wc_dict = get_sm_wcs(os.path.join(args.eosprefix, 'share/eos', 'parameters'))
    # convert to EOS parameters
    eos_dict = wcxf2eos(wc, sm_wc_dict)
    yaml.dump(eos_dict, f, default_flow_style=False)
    f.close()
    return 0


def smeftsim():
    from wcxf.converters.smeftsim import initialize_smeftsim_card, smeftsim_card_fill, smeftsim_card_text
    parser = argparse.ArgumentParser(description="""Command line script to convert a WCxf file to a MadGraph param_card file for SMEFTsim.""",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("FILE", nargs='?', help="Input file. Must be specified.",
                        type=argparse.FileType('r'), default="{}")
    parser.add_argument("--output", nargs='?', help="Output file. Default is wcxf2smeftsim_param_card.dat.", default="wcxf2smeftsim_param_card.dat")
    parser.add_argument("--input-scheme", nargs='?', help="Input parameters set. Can be either alpha (alpha_ew, m_Z, G_F) or mw (m_W, m_Z, G_F). Default is alpha.", choices=['alpha','mw'], default='alpha')
    parser.add_argument("--cutoff-scale", nargs='?', help="Value of the EFT cutoff scale in GeV. Default is 1 TeV.", type=float, default=1000)
    parser.add_argument("--model-set", nargs='?', help="SMEFTsim model set to be used. Can be either A or B, default is A.", choices=['A','B'],default="A")

    args = parser.parse_args()


    # read in & validate WCxf file
    wc = wcxf.WC.load(args.FILE)
    wc.validate()
    # check that the input is in the Warsaw mass basis. quit otherwise.
    if wc.basis != "Warsaw mass":
      print('''
WARNING: The input file must be in the 'Warsaw mass' basis!
Please translate into 'Warsaw mass' before converting to SMEFTsim.

Press 'i' to ignore this warning or any other key to exit.''')

      key = input()
      if key != 'i' and key != 'I': quit()

    f = open(args.output, 'w')

    # initialize and fill the dictionary for the param_card
    card = initialize_smeftsim_card(args.model_set)
    card_filled = smeftsim_card_fill(card, wc, args.model_set, args.cutoff_scale, args.input_scheme)

    # write output file
    f.write(smeftsim_card_text(args.model_set, args.input_scheme)[0])
    pylha.dump(card_filled, fmt='lha', stream = f)
    f.write(smeftsim_card_text(args.model_set, args.input_scheme)[1])
    f.close()
    return 0


def wcxf2dsixtools():
    from wcxf.converters import dsixtools
    parser = argparse.ArgumentParser(description="""Command line script to convert a WCxf file to a DsixTools Wilson coefficient file.""",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("FILE", nargs='?', help="Input file. If \"-\", read from standard input",
                        type=argparse.FileType('r'), default=sys.stdin)
    parser.add_argument("--output", nargs='?', help="Output file. If absent, print to standard output",
                        type=argparse.FileType('w'), default=sys.stdout)
    args = parser.parse_args()
    wc = wcxf.WC.load(args.FILE)
    wc.validate()
    dsixtools.wcxf2dsixtools(wc, stream=args.output)
    return 0


def dsixtools2wcxf():
    from wcxf.converters import dsixtools
    parser = argparse.ArgumentParser(description="""Command line script to convert DsixTools output files to a WCxf file.""",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("FILE", nargs='+', help="Input file(s).",
                        type=argparse.FileType('r'), default=sys.stdin)
    parser.add_argument("--output", nargs='?', help="Output file. If absent, print to standard output",
                        type=argparse.FileType('w'), default=sys.stdout)
    args = parser.parse_args()
    dsixtools.dsixtools2wcxf(tuple(f for f in args.FILE), stream=args.output)
    return 0



import unittest
import wcxf
from wcxf.converters import dsixtools
import os
import subprocess


my_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(my_path, '..', 'data')


with open(os.path.join(data_path, 'WCsInput-CPV-SMEFT.dat'), 'r') as f:
    wcin_lha = f.read()
with open(os.path.join(data_path, 'Options.dat'), 'r') as f:
    options = f.read()
with open(os.path.join(data_path, 'SMInput-CPV.dat'), 'r') as f:
    smin = f.read()
with open(os.path.join(data_path, 'WCsInput-CPV-SMEFT.json'), 'r') as f:
    wcin_json = f.read()
with open(os.path.join(data_path, 'WCsInput-CPV-SMEFT.yaml'), 'r') as f:
    wcin_yaml = f.read()


class TestDsixTools(unittest.TestCase):
    def test_smeftio(self):
        smeftio = dsixtools.SMEFTio()
        smeftio.load_wcxf(wcin_json)
        smeftio.load_wcxf(wcin_yaml)
        smeftio.load_initial((wcin_lha, options, smin))

    def test_wcxf2dsixtools(self):
        wc = wcxf.WC.load(wcin_json)
        d1 = dsixtools.load(dsixtools.wcxf2dsixtools(wc))
        smeftio = dsixtools.SMEFTio()
        smeftio.load_initial((wcin_lha, options, smin))
        d2 = dsixtools.load(smeftio.dump(smeftio.C_in))
        for key in ['SCALES', 'OPTIONS']:
            d1['BLOCK'].pop(key, None)
            d2['BLOCK'].pop(key, None)
        self.assertDictEqual(d1, d2)

    def test_dsixtools2wcxf(self):
        smeftio = dsixtools.SMEFTio()
        smeftio.load_initial((wcin_lha, options, smin))
        d1 = wcxf.WC.load(dsixtools.dsixtools2wcxf((wcin_lha, options, smin))).dict
        d2 = wcxf.WC.load(wcin_yaml).dict
        self.assertTrue(d1)
        self.assertTrue(d2)
        self.assertEqual(set(d1.keys()), set(d2.keys()))
        for k in d1:
            self.assertEqual(d1[k], d2[k], msg="Failed for {}".format(k))

    def test_cli_wcxf2dsixtools(self):
        smeftio = dsixtools.SMEFTio()
        smeftio.load_initial((wcin_lha, options, smin))
        d1 = dsixtools.load(smeftio.dump(smeftio.C_in))
        res = subprocess.run(['wcxf2dsixtools', '-'],
                             input=wcin_json.encode(),
                             stdout=subprocess.PIPE)
        s = res.stdout.decode('utf-8')
        d2 = dsixtools.load(s)
        self.assertDictEqual(d1, d2)

    def test_cli_dsixtools2wcxf(self):
        d1 = wcxf.WC.load(wcin_yaml).dict
        res = subprocess.run(['dsixtools2wcxf',
                             os.path.join(data_path, 'SMInput-CPV.dat'),
                             os.path.join(data_path, 'WCsInput-CPV-SMEFT.dat'),
                             os.path.join(data_path, 'Options.dat')],
                             stdout=subprocess.PIPE)
        s = res.stdout.decode('utf-8')
        d2 = wcxf.WC.load(s).dict
        self.assertDictEqual(d1, d2)



import json
import yaml
import logging
from collections import OrderedDict, Counter
import tempfile
import shutil
import os
import subprocess
from pandas import DataFrame

# the following is necessary to get pretty representations of
# OrderedDict and defaultdict instances in YAML
def _represent_dict_order(self, data):
    return self.represent_mapping('tag:yaml.org,2002:map', data.items())
yaml.add_representer(OrderedDict, _represent_dict_order)

# the following is necessary to have Null values be represented by
# emptyness rather than 'null'
def represent_none(self, _):
    return self.represent_scalar('tag:yaml.org,2002:null', '')
yaml.add_representer(type(None), represent_none)

# the following is necessary to load YAML mappings as OrderedDicts
_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG
def _dict_constructor(loader, node):
    return OrderedDict(loader.construct_pairs(node))
yaml.add_constructor(_mapping_tag, _dict_constructor)

def _load_yaml_json(stream, **kwargs):
    """Load a JSON or YAML file from a string or stream."""
    if isinstance(stream, str):
        ss = stream
    else:
        ss = stream.read()
    try:
        return json.loads(ss, **kwargs)
    except ValueError:
        return yaml.load(ss, **kwargs)

def _dump_json(d, stream=None, **kwargs):
    """Dump to a JSON string (if `stream` is None) or stream."""
    if stream is not None:
        return json.dump(d, stream, **kwargs)
    else:
        return json.dumps(d, **kwargs)

def _yaml_to_json(stream_in, stream_out, **kwargs):
    d = yaml.load(stream_in)
    return _dump_json(d, stream_out, **kwargs)

def _json_to_yaml(stream_in, stream_out, **kwargs):
    d = json.load(stream_in)
    return yaml.dump(d, stream_out, **kwargs)


def _testtex(s, delete=True):
    """Function that takes a string and tries to compile a LaTeX file with
    the string as document body. Returns a dictionary with keys 'success'
    (True or False) and 'log', trying to extract the first error message."""
    _preamble = r"""\documentclass{article}
    \usepackage{amsmath,amssymb}
    \begin{document}
    """
    _enddoc = r"""
    \end{document}"""
    doc = _preamble + s + _enddoc
    tmpd = tempfile.mkdtemp()
    tmpf = os.path.join(tmpd, 'textest.tex')
    with open(tmpf, 'w') as f:
        f.write(doc)
    try:
        p = subprocess.run(['latex', '-halt-on-error',
                            '-output-directory', tmpd, tmpf],
                           stdout=subprocess.DEVNULL)
    except FileNotFoundError:
        logging.warn('latex executable not found. Cannot check tex code')
        return {'success': True}
    if p.returncode == 0:
        res = {'success': True}
    else:
        res = {'success': False}
    logf = os.path.join(tmpd, 'textest.log')
    res['log'] = ''
    try:
        with open(logf, 'r') as f:
            logl = f.readlines()
    except FileNotFoundError:
        pass
    fail = False
    for i, l in enumerate(logl):
        if 'Undefined control sequence' in l or 'Emergency stop' in l:
            fail = True
            break
    if fail:
        l = logl[i]
        while l.strip() != '':
            res['log'] += logl[i]
            i += 1
            try:
                l = logl[i]
            except IndexError:
                break
    if delete:
        shutil.rmtree(tmpd)
    return res


class NamedInstanceMetaclass(type):
    # this is just needed to implement the getitem method on NamedInstanceClass
    # to allow the syntax MyClass['instancename'] as shorthand for
    # MyClass.get_instance('instancename'); same for
    # del MyClass['instancename'] instead of MyClass.del_instance('instancename')
    def __getitem__(cls, item):
        return cls.get_instance(item)

    def __delitem__(cls, item):
        return cls.del_instance(item)

class NamedInstanceClass(object, metaclass=NamedInstanceMetaclass):
    """Base class for classes that have named instances that can be accessed
    by their name.

    Parameters
    ----------
     - name: string

    Methods
    -------
     - del_instance(name)
         Delete an instance
     - get_instance(name)
         Get an instance
     - set_description(description)
         Set the description
    """

    def __init__(self, _name):
        if not hasattr(self.__class__, 'instances'):
            self.__class__.instances = OrderedDict()
        self.__class__.instances[_name] = self
        self._name = _name

    @classmethod
    def get_instance(cls, _name):
        return cls.instances[_name]

    @classmethod
    def del_instance(cls, _name):
        del cls.instances[_name]

    @classmethod
    def clear_all(cls):
        """Delete all instances."""
        cls.instances = OrderedDict()


class WCxf(object):
    """Base class for WCxf files (not meant to be used directly)."""

    @classmethod
    def load(cls, stream, **kwargs):
        """Load the object data from a JSON or YAML file."""
        wcxf = _load_yaml_json(stream, **kwargs)
        return cls(**wcxf)

    def dump(self, stream=None, fmt='json', **kwargs):
        """Dump the object data to a JSON or YAML file.

        Optional arguments:

        - `stream`: if None (default), return a string. Otherwise,
          should be a writable file-like object
        - `fmt`: format, should be 'json' (default) or 'yaml'

        Additional keyword arguments will be passed to the `json.dump(s)`
        or `yaml.dump` methods.
        """
        d = {k: v for k,v in self.__dict__.items() if k[0] != '_'}
        if fmt.lower() == 'json':
            # set indent=2 unless specified otherwise
            indent = kwargs.pop('indent', 2)
            return _dump_json(d, stream=stream,
                              indent=indent,
                              **kwargs)
        elif fmt.lower() == 'yaml':
            # set default_flow_style=False unless specified otherwise
            default_flow_style = kwargs.pop('default_flow_style', False)
            return yaml.dump(d, stream,
                             default_flow_style=default_flow_style,
                            **kwargs)
        else:
            raise ValueError("Format {} unknown: use 'json' or 'yaml'.".format(fmt))

class EFT(WCxf, NamedInstanceClass):
    """Class representing EFT files."""
    def __init__(self, eft, sectors, **kwargs):
        """Instantiate the EFT file object."""
        self.eft = eft
        self.sectors = sectors
        super().__init__(eft)
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def known_bases(self):
        """Return a list of known bases for this EFT."""
        return tuple(basis[1] for basis in Basis.instances if basis[0] == self.eft)

    @property
    def known_translators(self):
        """Return a list of known translators between bases of this EFT."""
        return tuple(t for t in Translator.instances if t[0] == self.eft)


class Basis(WCxf, NamedInstanceClass):
    """Class representing basis files."""
    def __init__(self, eft, basis, sectors, **kwargs):
        """Instantiate the basis file object."""
        self.eft = eft
        self.basis = basis
        super().__init__((eft, basis))
        for k, v in kwargs.items():
            setattr(self, k, v)
        if hasattr(self, 'parent'):
            try:
                self.sectors = Basis[self.eft, self.parent].sectors.copy()
                self.sectors.update(sectors)
            except (AttributeError, KeyError):
                raise ValueError("Parent basis {} not found".format(self.parent))
        else:
            self.sectors = sectors
        self._all_wcs = None

    @property
    def known_translators(self):
        """Return a list of known translators to and from this basis."""
        kt = {}
        kt['from'] = tuple(t for t in Translator.instances
                           if t[0] == self.eft and t[1] == self.basis)
        kt['to'] = tuple(t for t in Translator.instances
                         if t[0] == self.eft and t[2] == self.basis)
        return kt

    @property
    def all_wcs(self):
        """Return a list with all Wilson coefficients defined in this basis."""
        if self._all_wcs is None:
            self._all_wcs = [wc for sector, wcs in self.sectors.items() for wc in wcs]
        return self._all_wcs

    def validate(self):
        """Validate the basis file."""
        try:
            eft_instance = EFT[self.eft]
        except (AttributeError, KeyError):
            raise ValueError("EFT {} not defined".format(self.eft))
        unknown_sectors = set(self.sectors.keys()) - set(eft_instance.sectors.keys())
        if unknown_sectors:
            raise ValueError("Unknown sectors: {}".format(unknown_sectors))
        all_keys = [k for s in self.sectors.values() for k, v in s.items()]
        if len(all_keys) != len(set(all_keys)):  # we have duplicate keys!
            cnt = Counter(all_keys)
            dupes = [k for k, v in cnt.items() if v > 1]
            raise ValueError("Duplicate coefficients in different sectors:"
                             " {}".format(dupes))
        # check for LaTeX errors
        # string with all tex values
        alltex = '${}$'.format('$, $'.join([d.get('tex', '')
                                            for c in self.sectors.values()
                                            for d in c.values()
                                            if d is not None]))
        res = _testtex(alltex)
        if not res['success']:
            raise ValueError("Validation of basis {}/{}: "
                             .format(self.eft, self.basis)
                             + "LaTeX compilation errors encountered:\n"
                             + "{}".format(res['log']))

    def __repr__(self):
        return "wcxf.Basis('{}', '{}', {{...}})".format(self.eft, self.basis)

    def _repr_markdown_(self):
        md = "# Basis `{}` (EFT `{}`)\n\n".format(self.basis, self.eft)
        if hasattr(self, 'metadata') and 'description' in self.metadata:
            md += self.metadata['description'] + "\n\n"
        return md

    def _markdown_tables(self):
        md = "## Sectors\n\n"
        md += "The effective Lagrangian is defined as\n"
        # general definition of Leff according to WCxf papaer
        md += (r"$$\mathcal L_\text{eff} = -\mathcal H_\text{eff} ="
               r"\sum_{O_i= O_i^\dagger} C_i \, O_i + \sum_{O_i\neq O_i^\dagger}"
               r"\left( C_i \, O_i + C^*_i \, O^\dagger_i\right).$$")
        md += "\n\n"
        for s, wcs in self.sectors.items():
            md += "### `{}`\n\n".format(s)
            if wcs:
                md += "| WC name | Operator | Type |\n"
                # NB: this is meant for pandoc; it computes column widths
                # in latex by counting the number of "-" separators as
                # fractions of line width (default: 72)
                md += "|" + 18 * "-" + "|" + 48 * "-" + "|" + 6 * "-" + "|\n"
                for name, d in wcs.items():
                    if 'real' not in d or not d['real']:
                        t = 'C'
                    else:
                        t = 'R'
                    md += "| `{}` | ${}$ | {} |\n".format(name,
                                                          d.get('tex', ''), t)
        return md

    def __str__(self):
        md = self._repr_markdown_()
        md += self._markdown_tables()
        return md


class WC(WCxf):
    """Class representing Wilson coefficient files."""
    def __init__(self, eft, basis, scale, values, **kwargs):
        """Instantiate the Wilson coefficient file object."""
        self.eft = eft
        self.basis = basis
        self.scale = float(scale)
        self.values = values
        self._dict = None
        self._df = None
        super().__init__()
        for k, v in kwargs.items():
            setattr(self, k, v)

    @staticmethod
    def _to_number(v):
        """Turn a Wilson coefficient value - that could be a number or a Re/Im
        dict - into a number."""
        if isinstance(v, dict):
            return float(v.get('Re', 0)) + 1j*float(v.get('Im', 0))
        else:
            return float(v)

    @staticmethod
    def _to_complex_dict(v):
        """Turn a numeric Wilson coefficient value into a Re/Im dict if it is
        complex."""
        if v.imag != 0:
            return {'Re': float(v.real), 'Im': float(v.imag)}
        else:
            return float(v.real)

    @classmethod
    def dict2values(cls, d):
        return {k: cls._to_complex_dict(v) for k, v in d.items()}

    def __getitem__(self, key):
        try:
            return self.dict[key]
        except KeyError:
            return 0

    def validate(self):
        """Validate the Wilson coefficient file."""
        try:
            eft_instance = EFT[self.eft]
        except (AttributeError, KeyError):
            raise ValueError("EFT {} not defined".format(self.eft))
        try:
            basis_instance = Basis[self.eft, self.basis]
        except (AttributeError, KeyError):
            raise ValueError("Basis {} not defined for EFT {}".format(self.basis, self.eft))
        unknown_keys = set(self.values.keys()) - set(basis_instance.all_wcs)
        assert unknown_keys == set(), \
            "Wilson coefficients do not exist in this basis: " + str(unknown_keys)

    @property
    def dict(self):
        """Return a dictionary with the Wilson coefficient values.
        The dictionary will be cached when called for the first time."""
        if self._dict is None:
            self._dict = {k: self._to_number(v) for k, v in self.values.items()}
        return self._dict

    @property
    def df(self):
        """Return a pandas.DataFrame with the Wilson coefficient values
        split by real and imaginary part.
        The DataFrame will be cached when called for the first time."""
        if self._df is None:
            re = [self.dict[k].real for k in self.values]
            im = [self.dict[k].imag for k in self.values]
            self._df = DataFrame({'Re': re, 'Im': im},
                                 index=self.values,
                                 columns=('Re', 'Im'))
        return self._df

    def __repr__(self):
        return ("wcxf.WC(eft='{}', basis='{}', scale='{}', values={{...}})"
                .format(self.eft, self.basis, self.scale))

    def __str__(self):
        return self._repr_markdown_()

    def _repr_markdown_(self):
        md = "## WCxf Wilson coefficients\n\n"
        md += "**EFT:** `{}`\n\n".format(self.eft)
        md += "**Basis:** `{}`\n\n".format(self.basis)
        md += "**Scale:** {} GeV\n\n".format(self.scale)
        md += "### Values\n\n"
        md += "| WC name | Value |\n"
        # NB: this is meant for pandoc; it computes column widths
        # in latex by counting the number of "-" separators as
        # fractions of line width (default: 72)
        md += "|" + 20 * "-" + "|" + 52 * "-" + "|\n"
        for name, v in self.dict.items():
            md += "| `{}` | {} |\n".format(name, v)
        return md

    def _repr_html_(self):
        html = "<h3>WCxf Wilson coefficients</h3>\n\n"
        html += """<table>
  <thead>
    <tr>
      <th>EFT</th>
      <th>Basis</th>
      <th>scale</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <td><code>{}</code></td>
      <td><code>{}</code></td>
      <td>{} GeV</td>
    </tr>
  </tbody>
</table>
""".format(self.eft, self.basis, self.scale)
        html += "<h4>Values</h4>\n\n"
        html += self.df._repr_html_()
        return html

    def translate(self, to_basis, parameters=None):
        """Translate the Wilson coefficients to a different basis.
        Returns a WC instance."""
        if to_basis == self.basis:
            return self  # nothing to do
        try:
            translator = Translator[self.eft, self.basis, to_basis]
        except (KeyError, AttributeError):
            raise ValueError("No translator from basis {} to {} found.".format(self.basis, to_basis))
        return translator.translate(self, parameters=parameters)

    def match(self, to_eft, to_basis, parameters=None):
        """Match the Wilson coefficients to a different EFT.
        Returns a WC instance."""
        if to_eft == self.eft and to_basis == self.basis:
            return self  # nothing to do
        try:
            matcher = Matcher[self.eft, self.basis, to_eft, to_basis]
        except (KeyError, AttributeError):
            raise ValueError("No matcher from EFT {} in basis {} to EFT {} in basis {} found.".format(self.eft, self.basis, to_eft, to_basis))
        return matcher.match(self, parameters=parameters)


class Translator(NamedInstanceClass):
    """Class for translating between different bases of the same EFT."""
    def __init__(self, eft, from_basis, to_basis, function):
        """Initialize the Translator instance."""
        super().__init__((eft, from_basis, to_basis))
        self.eft = eft
        self.from_basis = from_basis
        self.to_basis = to_basis
        self.function = function

    def translate(self, WC_in, parameters=None):
        """Translate a WC object from `from_basis` to `to_basis`."""
        dict_out = self.function(WC_in.dict, WC_in.scale, parameters)
        # filter out zero values
        dict_out = {k: v for k, v in dict_out.items() if v != 0}
        values = WC.dict2values(dict_out)
        WC_out = WC(self.eft, self.to_basis, WC_in.scale, values)
        return WC_out


class Matcher(NamedInstanceClass):
    """Class for matching from a UV to an IR EFT."""
    def __init__(self, from_eft, from_basis, to_eft, to_basis, function):
        """Initialize the Matcher instance."""
        super().__init__((from_eft, from_basis, to_eft, to_basis))
        self.from_eft = from_eft
        self.from_basis = from_basis
        self.to_eft = to_eft
        self.to_basis = to_basis
        self.function = function

    def match(self, WC_in, parameters=None):
        """Translate a WC object in EFT `from_eft` and basis `from_basis`
        to EFT `to_eft` and basis `to_basis`."""
        dict_out = self.function(WC_in.dict, WC_in.scale, parameters)
        # filter out zero values
        dict_out = {k: v for k, v in dict_out.items() if v != 0}
        values = WC.dict2values(dict_out)
        WC_out = WC(self.to_eft, self.to_basis, WC_in.scale, values)
        return WC_out

def parametrized(dec):
    """Decorator for a decorator allowing it to take arguments.
    See https://stackoverflow.com/a/26151604."""
    def layer(*args, **kwargs):
        def repl(f):
            return dec(f, *args, **kwargs)
        return repl
    return layer

@parametrized
def translator(func, eft, from_basis, to_basis):
    """Decorator for basis translation functions.

    Usage:

    ```python
    @translator('myEFT', 'myBasis_from', 'myBasis_to')
    def myFunction(wc_dict_from):
        ... # do something
        return wc_dict_to
    ```
    """
    Translator(eft, from_basis, to_basis, func)
    return func


@parametrized
def matcher(func, from_eft, from_basis, to_eft, to_basis):
    """Decorator for matching functions.

    Usage:

    ```python
    @matcher('myEFT_from', 'myBasis_from', 'myEFT_to', 'myBasis_to')
    def myFunction(wc_dict_from):
        ... # do something
        return wc_dict_to
    ```
    """
    Matcher(from_eft, from_basis, to_eft, to_basis, func)
    return func
